% RunCulture.m -> GrowCell() grows the co-culture(models)
%     with the FBA assumption(Max growth, MOMA, ...)
%     in the defined mode of cultivation (batch, chemostat ...)
%     over time, and obtains a time dependent profiles
%     of the environmental changes (coculture)


function [coculture models] = GrowCell(params, degradation , death_release, models, coculture, totaltime, delta_t, time_steps , MM  )

    GLPKERROR = 1e-6;
    Flow = coculture.tankflow;
    Volume = coculture.volume;
    Dilution = Flow/Volume;

    vgro_culture=ones(1, length(coculture.species))    ;



    t=0;  % used to make sure program stops at totaltime:
    i=0;  % used to index the time point:
    while( any(vgro_culture)  && (t<totaltime))
        i=i+1;

        if (isfield(params,'verbose') && (params.verbose == 1))
            disp([' ==== Time (0:' num2str(totaltime) ' hr): '  num2str(t) ' hr ====']);
        end

        % (1) Update Lb
        CellDensity = sum(coculture.bio(:,i)); %[gDW of cell/L]

        % ############# QSSA method: scaled concentrations ############
        % QSSA method assumes a evenly distributed environment
        % that gives you an idea how much glc one cell can 'own'
        % Each cell couldn't consume glucose 'own' by other cells

        % 'Feel' scaled ctns at t(i) to determine uptake rate at t(i)
        Xctns = coculture.ctns(:,i)';

        %Unit: mmol     mmol  /  gDW of cell * hr
        %      ----   = -----/  -----------------
        %      gDW hr     L /        L
        % Max uptake by the ctns per cell can 'feel'
        Xctns_scaled = Xctns/(CellDensity*delta_t);

        % Max uptake by Michaelis-Menten
        r_metX = MM.Vmax  .* Xctns./( Xctns + MM.km ) ;

        % Take the conservative one (closer to zero)
        % This will be used to update Lb for each EX_rxns
        EX_metX_lb = min(r_metX , Xctns_scaled);

        % disp(['Initial glucose supply rate = ' num2str(EX_metX_lb(9),3) , '[mmol/gDW-hr]'] )

        % Update lb(mets_j, species_k) for each Exchangable mets j in species k
        for k=1:length(coculture.species)
            % matrix of secretable mets' lb
            clear EX_lb;
            EX_lb = EX_metX_lb(   models(k).EXmetsMEDIUMid   );
            %EX_lb_name is only used to check what lb is updated
            EX_lb_name = coculture.mets(  models(k).EXmetsMEDIUMid  );

            models(k).lb(   models(k).EXrxnsID  )  = -1*EX_lb  ;
            % keep a history record
            models(k).lbhistory(:,i) = models(k).lb;
        end

        % ############# QSSA method: scaled concentrations ############

        % (2) Run FBA
        for k=1:length(coculture.species)
            if (coculture.bio(k,i) > 0)
                v_tmp = zeros(length(models(k).rxns) , 1);
                v_species = zeros(length(models(k).rxns) , 1);
                % (A)do simple FBA that maximize growth
                if params.growthassumption ==1
                    [v_species, MOMAobj_sol, status] = MOMA_AV(models(k), models(k).v_WT_msf);
                else
                    %Always minimize sum of total flux
                    EX = find(models(k).isEX);
                    L = find(models(k).lb<0);
                    nu = intersect(EX,L);

                    [v_tmp, v_species, activity, status] = FBA_min_fluxAV(models(k), params.verbose);

                    %check for potential glpk bug
                    x = intersect(models(k).EXmetsID, find((v_species - models(k).lb) < -1e-6) ) ;
                    if(~isempty(x))
                        disp(['In ' char(models(k))  'Time step ' num2str(i)])
                        error('Flux solution violates LB too much')
                    end
                end

                %(1)set small flux: |v|<1e-6 to zero,
                %(2)zero growth -> zero fluxes
                id_0 = find( abs(v_species) < GLPKERROR );
                v_species( id_0) = 0;
                if (v_species(  find(models(k).c)  ) < GLPKERROR )
                    v_species = zeros(size(v_species));
                end

                % (B)Store results
                % Record FBA derived growth and fluxes,
                % Note that v act on the time "interval" so if there are
                % 201 time points, there are only 200 fluxes
                models(k).FBA_EXflux(:,i) = v_species( models(k).EXrxnsID  );
                models(k).mu(i) = v_species( find(models(k).c)  );
                models(k).allflux = v_species;
                models(k).glpkstatus = status;
                coculture.allflux{k}(:,i) =  v_species;

                % These are for debugging purpose, but they are useful!
                coculture.absfluxactivity{k}(i) = sum(abs(v_species));
                coculture.glpkstatus{k}(i) = models(k).glpkstatus;
                coculture.lb{k}(:,i) = models(k).lb;
                coculture.ub{k}(:,i) = models(k).ub;
                coculture.glpkflux{k}(:,i) =  v_tmp;
                % These are for debugging purpose, but they are useful!

                %variable used to stop calculating growth in batch mode
                %if FBA is not performed this will give zero growth
                if (params.cultivation == 0)
                    vgro_culture(k) = v_species( find(models(k).c)  );
                else
                    vgro_culture(k) = 1;
                end

            else
                models(k).mu(i) = 0;
            end
        end

       % (3) Update environment

        %(3.1) Update BIOMASS
        for k=1:length(coculture.species)
            %Currently, death_rate = 0
            mu_net = models(k).mu(i) - models(k).death_rate ;

            %exponential growth equation
            postgrowth = coculture.bio(k,i)*exp(mu_net*delta_t);
            coculture.bio(k,i+1) = postgrowth  ;
        end

        %(3.2) Update All metabolites

        %(3.2.A) FBA contribution (exponential):
        %    [X(t+1)]=[X(t)] + (V*[B(t)]/mu )*(exp(mu*dt)-1)
        X_next = coculture.ctns(:,i);
        for k=1:length(coculture.species)
            if (models(k).mu(i) > 0)
                dx_ex =[];  v=[]; b=[]; mu_net=[]; total_fba_v=[];

                % size of m, all metabolites
                dx_ex = -1*models(k).S*(models(k).allflux .* models(k).isEX); % dX of all mets
                % Eliminate rounding error by setting small ctns changes
                % to be exactly zero                           2012/3/25
                dx_ex( find(abs(dx_ex) < GLPKERROR  ) ) = 0 ;

                v = dx_ex(models(k).EXmetsID);  % dX of only exchangeable mets

                b = coculture.bio(k,i);
                %mu_net = models(k).mu(i) - models(k).death_rate;
                mu_net = models(k).mu(i) - models(k).death_rate ;
                if (mu_net~=0)
                    %according to Varma paper, this has nothing to do with volume
                    total_fba_v = v*(b/mu_net)*(exp(mu_net*delta_t)-1);
                else
                    total_fba_v = zeros(size(v,1) , size(v,2)  );
                end

                %Sparse flux vector secreted into medium
                v2medium = zeros(size(X_next));
                v2medium(models(k).EXmetsMEDIUMid) = v;
                v2medium_time{:,i,k} = v2medium;

                %Sparse vector of DELTA_concentration
                fba2medium = zeros(size(X_next));
                fba2medium(models(k).EXmetsMEDIUMid) = total_fba_v;

                X_next = X_next + fba2medium;
            else
                if (coculture.bio(k,1) > 0)
                    v2medium = zeros(size(X_next));
                    v2medium_time{:,i,k} = v2medium;
                end
            end

        end

            %(3.2.B) Take care of inflow and outflow:

        %(3.2.B-1)Batch mode, EBC paper by default
        if (params.cultivation == 0)
            Real_X_next = X_next;
        end

        %(3.2.B-2)Chemostat dilution, always dilute X_next, B_next


        if t < coculture.dietswitchtime
            chemo_MetCtnsIn = coculture.reservoir;
        else
            %Diet is switched to a new one
            disp(t);
            chemo_MetCtnsIn = coculture.newreservoir;
            error('no newreservoir, probably');
        end

        % Diet paper use chemostat, so will enter this cultivation == 1:
        if(params.cultivation == 1)
            %Biomass dilution
            for k=1:length(coculture.species)
                %B_next = coculture.bio(k,i)*exp(net_vgro*delta_t);
                B_next = coculture.bio(k,i+1);
                B_curr = coculture.bio(k,i);
                coculture.bio(k,i+1) = B_next - Dilution * B_curr * delta_t ;
                %coculture.bio(k,i+1) = B_next ;
            end

            %Metabolite dilution, note that Dilution = Flow/Volume
            X_curr = coculture.ctns(:,i);
            %debug variables
            %TMPTMP = (chemo_MetCtnsIn - X_curr);
            %Y = Dilution*(chemo_MetCtnsIn - X_curr)*delta_t ;
            Real_X_next = X_next + Dilution*(chemo_MetCtnsIn - X_curr)*delta_t ;
            %                 Real_X_next = X_next + Dilution*(chemo_MetCtnsIn )*delta_t ;
            %(3.2.B-3)serial transfer: add concentration at each given time point
            % %           elseif(params.cultivation == 2)
            % %                 a = int64(coculture.transfer_hr/delta_t);
            % %                 b = int64(totaltime/delta_t);
            % %                 i_tick = [1:a:b] ;
            % %                 if(t>0 && ismember(i,i_tick))
            % %                     ratio = coculture.transfer_ratio;
            % %
            % %                     %Biomass is transferred
            % %                     for k=1:length(coculture.species)
            % %                         coculture.bio(k,i+1) = ratio*coculture.bio(k,i+1);
            % %                     end
            % %                     %Concentration is
            % %                     %         transferred  + fresh medium
            % %                     Real_X_next = ratio*X_next + chemo_MetCtnsIn;
            % %                 end

            %(3.2.B-4)fed-batch: add concentration at each given time point
            % %            elseif(params.cultivation == 3)
            % %                  a = int64(coculture.transfer_hr/delta_t);
            % %                  b = int64(totaltime/delta_t);
            % %                  i_tick = [1:a:b] ;
            % %                 if(t>0 && ismember(i,i_tick))
            % %                     %Concentration is replenshed every
            % %                     Real_X_next = X_next + chemo_MetCtnsIn;
            % %                 end
        end

        %(3.2.C) Release from dead cell, (currently not implemented)
        % At present do not release from death cell 2012/2/1

        % Metabolites released from dead strain (The COSMO paper idea):
        Released_from_Bio = zeros(size(coculture.ctns(:,i),1)  , 1 );

        coculture.ctns(:,i+1) = Real_X_next + Released_from_Bio/Volume; %;

        %(4) Death , (currently not implemented):

        %(5) Avoid negative concentrations
        ind = find(coculture.ctns(:, i+1) <0);
        coculture.ctns(ind, i+1) =0;

        t = t + delta_t;
    end

    %Store results into coculture structure
    for k=1:length(coculture.species)
        
		coculture.mu_model{k} = models(k).mu;
		coculture.death_rate{k} = models(k).death_rate;
		coculture.finaltimestep(k) = length(find(coculture.mu_model{k}))+1 ;
		coculture.deadhour(k) = coculture.finaltimestep(k)*delta_t;
		coculture.ctns_timepoints(k) = coculture.finaltimestep(k) + 1 ;
		
		if (coculture.bio(k,1) > 0)
            for z=1:size(v2medium_time , 2)
                tmp(:,z) = v2medium_time{:,z,k};
            end

            coculture.v_model2medium{k} = tmp; % This is the V_ex
            
        else
			coculture.v_model2medium{k} = [];
		end
    end


end






